#pragma warning(disable:4244)
#pragma warning(disable:4018)

#include "definitions.h"
#include "textoutput.h"
#include "precisetimer.h"
#include "particle.h"

DclVector getWorld();

const int MAX_NUM_PARTICLES = 10000;

const int NUM_SPLIT_PARTICLES = 199;


// Integration step h
const DclFloat  TIMESTEP = 0.005;

// Elasticity of collision with walls
const DclFloat  WORLD_ELASTICITY = 0.6;



//////////// GLOBALS /////////////////
std::vector<Particle> vectorParticles;
int frame = 0;

//////////////////////////////////////

void  visualizeParticles();

DclFloat  colors[7][3] = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}, {1, 1, 0}, {1, 0, 1}, {0, 1, 1}, {1, 1, 1}};

//------------------------------------------------------------------------------------------------------------------------------------

void  createInitialParticle()
{
  vectorParticles.clear();
  Particle  particle;
  DclVector world = getWorld();

  // physical properties
  particle.position.x = 0;
  particle.position.y = -world.y / 2.0;
  particle.position.z = 0;

  particle.velocity.x = (rand() - RAND_MAX / 2.0) * 5.0 / (DclFloat)RAND_MAX;
  particle.velocity.y = 10.0 + rand() * 20.0 / (DclFloat)RAND_MAX;
  particle.velocity.z = (rand() - RAND_MAX / 2.0) * 5.0 / (DclFloat)RAND_MAX;

  particle.force.clear();

  particle.mass = 10.0;
    
  particle.integration = EulerCromer;
  particle.radius = 0.2;
  particle.r = particle.g = particle.b = 0.3;
  particle.liveCounter = 70;

  // put the particle into the vector
  vectorParticles.push_back(particle);
}

// initialize the particle properties
void  initializeSimulation()
{
  vectorParticles.reserve(MAX_NUM_PARTICLES);

  createInitialParticle();
}

//------------------------------------------------------------------------------------------------------------------------------------

// add external forces to the sum of forces acting on the particle
void  addForces(Particle& particle)
{
  particle.force += DclVector(0, 1.0, 0) * particle.mass * -9.81; // gravitational force
}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one time-integration step
void  integrate(Particle& particle)
{
  switch (particle.integration)
  {
    case Euler:
      particle.position = particle.position + particle.velocity * TIMESTEP;
      particle.velocity = particle.velocity + particle.force / particle.mass * TIMESTEP;
      break;

    case EulerCromer:
      particle.velocity = particle.velocity + particle.force / particle.mass * TIMESTEP;
      particle.position = particle.position + particle.velocity * TIMESTEP;
      break;

    default:
      break;
  }
}

//------------------------------------------------------------------------------------------------------------------------------------

// collision handling with the world
void  worldCollision(Particle& particle)
{
  DclVector world = getWorld();

  world *= 0.5;

  if (particle.position.x < -world.x || particle.position.x > world.x)
  {
    particle.velocity.x *= -WORLD_ELASTICITY;
    particle.position.x = SIGN(particle.position.x) * world.x;
  }

  if (particle.position.y < -world.y || particle.position.y > world.y)
  {
    particle.velocity.y *= -WORLD_ELASTICITY;
    particle.position.y = SIGN(particle.position.y) * world.y;
  }

  if (particle.position.z < -world.z || particle.position.z > world.z)
  {
    particle.velocity.z *= -WORLD_ELASTICITY;
    particle.position.z = SIGN(particle.position.z) * world.z;
  }
}

//------------------------------------------------------------------------------------------------------------------------------------

// compute and output the total energies in the scene
void  computeEnergies()
{
  DclFloat  e_kin = 0, e_pot = 0;
  DclVector world = getWorld();

  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    Particle& particle = vectorParticles[part];

    e_kin += particle.velocity.length2() * 0.5 * particle.mass;
    e_pot += (particle.position.y + world.z * 0.5) * particle.mass * 9.81;
  }

  Textout::printvar(6, 0, "Kinetic energy %f, Potential energy %f, Sum %f", e_kin, e_pot, e_kin + e_pot);
}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one simulation step
void  simulationStep()
{
  static int splitCounter = 0;

  // reset the vector of external forces
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    vectorParticles[part].force.clear();
  }

  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    Particle&  oldParticle = vectorParticles[part];

    oldParticle.liveCounter --;

    if (oldParticle.liveCounter == 1)
    {
      DclVector oldImpulse = oldParticle.velocity * oldParticle.mass;
      Textout::printvar(7, 0, "Impulse before: (%f, %f, %f)", oldImpulse.x, oldImpulse.y, oldImpulse.z);

      DclVector newImpulse(0, 0, 0);

      DclVector oldVelocity = oldParticle.velocity;

      for (int i = part + 1; i <= part + NUM_SPLIT_PARTICLES; i ++)
      {
        Particle  newParticle;
        newParticle = oldParticle;
        newParticle.velocity = oldVelocity;
        if (i % 2 == 1)
        {
          DclVector v(rand() - RAND_MAX / 2, rand() - RAND_MAX / 2, rand() - RAND_MAX / 2);
          v.normalize();
          v *= 5.0;

          newParticle.velocity += v;
          vectorParticles[i - 1].velocity -= v;
        }
        vectorParticles.push_back(newParticle);
      }

      int color =  splitCounter % 7;
      splitCounter ++;

      for (int i = part; i <= part + NUM_SPLIT_PARTICLES; i ++)
      {
        Particle &particle = vectorParticles[i];
        particle.mass /= (DclFloat)(NUM_SPLIT_PARTICLES + 1);
        // M = V * D --> V = M/D --> r^3 ~ V --> rold^3 * Vold = rnew^3 * Vnew --> rnew = sqrt3(rold^3 * Vnew / Vold)
        particle.radius = pow(particle.radius * particle.radius * particle.radius / (DclFloat)(NUM_SPLIT_PARTICLES + 1), 1.0 / 3.0);
        DclFloat fac = rand() / (DclFloat)RAND_MAX;
        particle.r = colors[color][0]* fac;
        particle.g = colors[color][1]* fac;
        particle.b = colors[color][2]* fac;

        newImpulse += particle.velocity * particle.mass;
      }
      Textout::printvar(8, 0, "Impulse after: (%f, %f, %f)", newImpulse.x, newImpulse.y, newImpulse.z);
    }
  }

  if (frame && (frame % 200) == 0)
    createInitialParticle();

  // compute external forces on the particles
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    addForces(vectorParticles[part]);
  }

  // compute collision impulses
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    worldCollision(vectorParticles[part]);
  }

  // perform one integration step
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    integrate(vectorParticles[part]);
  }

  // compute and output energies
  computeEnergies();

  // draw particles
  visualizeParticles();

  // measure time for frame
  double  dt = PreciseTimer::stop("SimulationStep");
  PreciseTimer::start("SimulationStep");

  // output statistics
  Textout::printvar(5, 0, "Frame %d, Frames/second %f, Num. particles %d", frame, 1000.0 / dt, vectorParticles.size());
  frame ++;
}


